
from logging import getLogger

from .args import FORCE, LIKE, FINISH, START, WORKER, parse_pairs, KARG
from ..squeal.tables.pipeline import PipelineType
from ..stoats.pipeline import run_pipeline

log = getLogger(__name__)


def statistics(args, db):
    '''
## statistics

    > ch2 statistics

Generate any missing statistics.

    > ch2 statistics --force [DATE]

Delete statistics after the date (or all, if omitted) and then generate new values.
    '''
    run_pipeline(db, PipelineType.STATISTIC,
                 force=args[FORCE], like=args[LIKE], start=args[START], finish=args[FINISH],
                 worker=args[WORKER] is not None, id=args[WORKER], **parse_pairs(args[KARG]))




from logging import getLogger

import pandas as pd

from .args import SUB_COMMAND, START, FINISH, NAMES, \
    OWNER, CONSTRAINT, SCHEDULE, SOURCE_IDS, STATISTIC_QUARTILES, PRINT, FORMAT, \
    CSV, DESCRIBE, MAX_COLUMNS, MAX_ROWS, WIDTH, MAX_COLWIDTH, TABLE, NAME, STATISTICS
from ..data import df, statistics, statistic_quartiles
from ..squeal import *
from ..squeal.support import Base

# to stop the import * from being deleted
StatisticName

log = getLogger(__name__)


def dump(args, db):
    '''
## dump

    > ch2 dump COMMAND

Simple access to the database - similar to the interface provided in Jupyter notebooks,
but accessed from the command line.

The format can be selected with `--print` (the default), `--csv` and `--describe`.

For full options see `ch2 data -h` and `ch2 data COMMAND -h`

### Examples

    > ch2 dump --csv table StatisticName

Will print the contents of the StatisticName table in CSV format.

    > ch2 dump statistics '%HR%' --constraint 'ActivityGroup "Bike"' --start 2018-01-01

Will print HR-related statistics from the start of 2018 for the given activity group.
    '''

    with db.session_context() as s:

        if args[SUB_COMMAND] == TABLE:
            name = args[NAME]
            if name in globals():
                table = globals()[name]
                if issubclass(table, Base):
                    frame = df(s.query(table))
                else:
                    raise Exception('%s is not a mapped table' % name)
            else:
                raise Exception('%s does not exist' % name)
        elif args[SUB_COMMAND] == STATISTICS:
            frame = statistics(s, *args[NAMES], start=args[START], finish=args[FINISH],
                               owner=args[OWNER], constraint=args[CONSTRAINT],
                               schedule=args[SCHEDULE], source_ids=args[SOURCE_IDS])
        elif args[SUB_COMMAND] == STATISTIC_QUARTILES:
            frame = statistic_quartiles(s, *args[NAMES], start=args[START], finish=args[FINISH],
                                        owner=args[OWNER], constraint=args[CONSTRAINT],
                                        schedule=args[SCHEDULE], source_ids=args[SOURCE_IDS])
        else:
            raise Exception('Unexpected %s: %s' % (SUB_COMMAND, args[SUB_COMMAND]))

        pd.options.display.max_columns = args[MAX_COLUMNS]
        if args[MAX_COLWIDTH] is not None: pd.options.display.max_colwidth = args[MAX_COLWIDTH]
        pd.options.display.max_rows = args[MAX_ROWS]
        pd.options.display.width = args[WIDTH]

        if args[FORMAT] == PRINT:
            print(frame)
        elif args[FORMAT] == CSV:
            print(frame.to_csv())
        elif args[FORMAT] == DESCRIBE:
            print(frame.describe(include='all'))



from logging import getLogger

from .climb import add_climb, CLIMB_CNAME
from .database import Counter, add_statistics, add_activity_group, add_activity_constant, add_topic, add_topic_field, \
    add_diary, add_activities, add_monitor, name_constant, add_nearby, add_constant, add_loader_support
from .impulse import add_impulse
from .power import add_power_estimate
from ..lib.schedule import Schedule
from ..sortem.file import SRTM1_DIR
from ..squeal.tables.statistic import StatisticJournalType
from ..squeal.tables.topic import TopicJournal
from ..squeal.types import short_cls
from ..stoats.calculate.activity import ActivityCalculator
from ..stoats.calculate.elevation import ElevationCalculator
from ..stoats.calculate.monitor import MonitorCalculator
from ..stoats.calculate.power import PowerCalculator
from ..stoats.calculate.segment import SegmentCalculator
from ..stoats.calculate.summary import SummaryCalculator
from ..stoats.display.activity import ActivityDiary
from ..stoats.display.impulse import ImpulseDiary
from ..stoats.display.monitor import MonitorDiary
from ..stoats.display.nearby import NearbyDiary
from ..stoats.display.segment import SegmentDiary
from ..stoats.names import BPM, FTHR, LONGITUDE, LATITUDE, HEART_RATE, SPEED, DISTANCE, ALTITUDE, DEG, MS, M, CADENCE, \
    RPM, FITNESS_D, FATIGUE_D
from ..stoats.read.monitor import MonitorReader
from ..stoats.read.segment import SegmentReader
from ..uweird.fields.topic import Text, Float, Score0

log = getLogger(__name__)


def default(db, no_diary=False):

    with db.session_context() as s:

        # the following users helper functions (add_...) but you can also
        # execute arbitrary python code, use the session, etc.

        add_loader_support(s)

        # basic activities

        c = Counter()
        bike = add_activity_group(s, 'Bike', c, description='All cycling activities')
        run = add_activity_group(s, 'Run', c, description='All running activities')
        # sport_to_activity maps from the FIT sport field to the activity defined above
        add_activities(s, SegmentReader, c,
                       owner_out=short_cls(SegmentReader),
                       sport_to_activity={'cycling': bike.name, 'running': run.name},
                       record_to_db={'position_lat': (LATITUDE, DEG, StatisticJournalType.FLOAT),
                                     'position_long': (LONGITUDE, DEG, StatisticJournalType.FLOAT),
                                     'heart_rate': (HEART_RATE, BPM, StatisticJournalType.INTEGER),
                                     'enhanced_speed': (SPEED, MS, StatisticJournalType.FLOAT),
                                     'distance': (DISTANCE, M, StatisticJournalType.FLOAT),
                                     'enhanced_altitude': (ALTITUDE, M, StatisticJournalType.FLOAT),
                                     'cadence': (CADENCE, RPM, StatisticJournalType.INTEGER)})

        # statistics pipeline (called to calculate missing statistics)

        # FF-model parameters
        # 7 and 42 days as for training peaks
        # https://www.trainingpeaks.com/blog/the-science-of-the-performance-manager/
        fitness=((42, 1), (260, 1/6))
        fatigue=((7, 5),)

        c = Counter()
        add_statistics(s, ElevationCalculator, c, owner_in='[unused - data via activity_statistics]')
        add_power_estimate(s, c, bike, vary='')
        add_climb(s, bike)
        add_impulse(s, c, bike, fitness=fitness, fatigue=fatigue)
        add_statistics(s, ActivityCalculator, c,
                       owner_in=f'{short_cls(SegmentReader)},{short_cls(PowerCalculator)},{short_cls(ElevationCalculator)}',
                       climb=name_constant(CLIMB_CNAME, bike))
        add_statistics(s, SegmentCalculator, c, owner_in=short_cls(SegmentReader))
        add_statistics(s, MonitorCalculator, c, owner_in=short_cls(MonitorReader))

        # need to call normalize here because schedule isn't a schedule type column,
        # but part of a kargs JSON blob.
        # also, add year first so that monthly doesn't get confused by extra stats range
        add_statistics(s, SummaryCalculator, c, schedule=Schedule.normalize('y'))
        add_statistics(s, SummaryCalculator, c, schedule=Schedule.normalize('m'))

        # obviously you need to edit these parameters - see `ch2 constants Nearby.Bike`
        add_nearby(s, c, bike, 'Santiago', -33.4, -70.4, fraction=0.1, border=150)

        # diary pipeline (called to display data in the diary)

        c = Counter()
        add_diary(s, MonitorDiary, c)
        # these tie-in to the constants used in add_impulse()
        add_diary(s, ImpulseDiary, c,
                  fitness=[name_constant(FITNESS_D % days, bike) for (days, _) in fitness],
                  fatigue=[name_constant(FATIGUE_D % days, bike) for (days, _) in fatigue])
        add_diary(s, ActivityDiary, c)
        add_diary(s, SegmentDiary, c)
        add_diary(s, NearbyDiary, c)

        # monitor pipeline

        c = Counter()
        add_monitor(s, MonitorReader, c)

        # constants used by statistics

        add_activity_constant(s, bike, FTHR,
                              description='Heart rate at functional threshold (cycling). See https://www.britishcycling.org.uk/knowledge/article/izn20140808-Understanding-Intensity-2--Heart-Rate-0',
                              units=BPM, statistic_journal_type=StatisticJournalType.INTEGER)
        add_activity_constant(s, run, FTHR,
                              description='Heart rate at functional threshold (running).',
                              units=BPM, statistic_journal_type=StatisticJournalType.INTEGER)
        add_constant(s, SRTM1_DIR, description='Directory containing STRM1 hgt files for elevations (see http://dwtkns.com/srtm30m)',
                     single=True, statistic_journal_type=StatisticJournalType.TEXT)

        if not no_diary:

            # a basic diary

            c = Counter()
            diary = add_topic(s, 'Diary', c)
            add_topic_field(s, diary, 'Notes', c, StatisticJournalType.TEXT,
                            display_cls=Text)
            # now provided via monitor
            # add_topic_field(s, diary, 'Rest HR', c,
            #                 units=BPM, summary='[avg]',
            #                 display_cls=Integer, lo=25, hi=75)
            add_topic_field(s, diary, 'Weight', c, StatisticJournalType.FLOAT,
                            units='kg', summary='[avg],[msr]',
                            display_cls=Float, lo=50, hi=100, dp=1)
            add_topic_field(s, diary, 'Sleep', c, StatisticJournalType.FLOAT,
                            units='h', summary='[avg]',
                            display_cls=Float, lo=0, hi=24, dp=1)
            add_topic_field(s, diary, 'Mood', c, StatisticJournalType.FLOAT,
                            summary='[avg]',
                            display_cls=Score0)
            add_topic_field(s, diary, 'Nutrition', c, StatisticJournalType.TEXT,
                            summary='[cnt]',
                            display_cls=Text)
            add_topic_field(s, diary, 'Soreness', c, StatisticJournalType.TEXT,
                            summary='[cnt]',
                            display_cls=Text)
            add_topic_field(s, diary, 'Medication', c, StatisticJournalType.TEXT,
                            summary='[cnt]',
                            display_cls=Text)
            add_topic_field(s, diary, 'Weather', c, StatisticJournalType.TEXT,
                            summary='[cnt]',
                            display_cls=Text)
            add_topic_field(s, diary, 'Route', c, StatisticJournalType.TEXT,
                            summary='[cnt]',
                            display_cls=Text)

        # finally, set the TZ so that first use of the diary doesn't wipe all our intervals
        TopicJournal.check_tz(log, s)



from argparse import ArgumentParser
from genericpath import exists
from logging import getLogger
from os import makedirs
from os.path import dirname, expanduser, realpath, normpath, relpath, join
from re import compile, sub
from typing import Mapping

from ..lib.date import to_date, to_time

log = getLogger(__name__)

CH2_VERSION = '0.19.12'

PROGNAME = 'ch2'
COMMAND = 'command'
TOPIC = 'topic'

ACTIVITIES = 'activities'
CONSTANTS = 'constants'
DEFAULT_CONFIG = 'default-config'
DIARY = 'diary'
DUMP = 'dump'
FIT = 'fit'
FIX_FIT = 'fix-fit'
GARMIN = 'garmin'
H, HELP = 'h', 'help'
NO_OP = 'no-op'
PACKAGE_FIT_PROFILE = 'package-fit-profile'
STATISTICS = 'statistics'
TEST_SCHEDULE = 'test-schedule'

ACTIVITY = 'activity'
ACTIVITY_GROUP = 'activity-group'
ACTIVITY_GROUPS = 'activity-groups'
ACTIVITY_JOURNALS = 'activity-journals'
ACTIVITY_JOURNAL_ID = 'activity-journal-id'
ADD_HEADER = 'add-header'
AFTER = 'after'
AFTER_BYTES = 'after-bytes'
AFTER_RECORDS = 'after-records'
ALL_MESSAGES = 'all-messages'
ALL_FIELDS = 'all-fields'
ARG = 'arg'
BORDER = 'border'
COMPACT = 'compact'
CONSTRAINT = 'constraint'
CONSTANT = 'constant'
CONTEXT = 'context'
CSV = 'csv'
D = 'd'
DATABASE = 'database'
DATE = 'date'
DELETE = 'delete'
DESCRIBE = 'describe'
DEV = 'dev'
DIR = 'dir'
DISCARD = 'discard'
DROP = 'drop'
F = 'f'
FAST = 'fast'
FIELD = 'field'
FIELDS = 'fields'
FINISH = 'finish'
FIX_CHECKSUM = 'fix-checksum'
FIX_HEADER = 'fix-header'
FORMAT = 'format'
FORCE = 'force'
FTHR = 'fthr'
GREP = 'grep'
GROUP = 'group'
HEADER_SIZE = 'header-size'
HEIGHT = 'height'
INTERNAL = 'internal'
JUPYTER = 'jupyter'
K = 'k'
KARG = 'karg'
LABEL = 'label'
LATITUDE = 'latitude'
LIKE = 'like'
LIMIT_BYTES = 'limit-bytes'
LIMIT_RECORDS = 'limit-records'
L, LOG = 'l', 'log'
LOGS = 'logs'
LONGITUDE = 'longitude'
LIST = 'list'
M, MESSAGE = 'm', 'message'
MATCH = 'match'
MAX_BACK_CNT = 'max-back-cnt'
MAX_COLUMNS = 'max-columns'
MAX_COLWIDTH = 'max-colwidth'
MAX_DROP_CNT = 'max-drop-cnt'
MAX_DELTA_T = 'max-delta-t'
MAX_FWD_LEN = 'max-fwd-len'
MAX_ROWS = 'max-rows'
MAX_RECORD_LEN = 'max-record-len'
MIN_SYNC_CNT = 'min-sync-cnt'
MONITOR = 'monitor'
MONITOR_JOURNALS = 'monitor-journals'
MONTH = 'month'
MONTHS = 'months'
NAME = 'name'
NAME_BAD = 'name-bad'
NAME_GOOD = 'name-good'
NAMES = 'names'
NOT = 'not'
O, OUTPUT = 'o', 'output'
OWNER = 'owner'
PASS = 'pass'
PATH = 'path'
PLAN = 'plan'
PRINT = 'print'
PROFILE_VERSION = 'profile-version'
PROTOCOL_VERSION = 'protocol-version'
PWD = 'pwd'
RAW = 'raw'
RECORDS = 'records'
ROOT = 'root'
RUN = 'run'
SEGMENT_JOURNALS = 'segment-journals'
SEGMENTS = 'segments'
SERVICE = 'service'
SET = 'set'
SCHEDULE = 'schedule'
SHOW = 'show'
SLICES = 'slices'
SOURCE_IDS = 'source-ids'
START = 'start'
STATISTIC_NAMES = 'statistic-names'
STATISTIC_JOURNALS = 'statistic-journals'
STATISTIC_QUARTILES = 'statistic-quartiles'
STATUS = 'status'
STOP = 'stop'
SUB_COMMAND = 'sub-command'
TABLE = 'table'
TABLES = 'tables'
TOKENS = 'tokens'
UNLOCK = 'unlock'
USER = 'user'
VALIDATE = 'validate'
V, VERBOSITY = 'v', 'verbosity'
VALUE = 'value'
VERSION = 'version'
W, WARN = 'w', 'warn'
WAYPOINTS = 'waypoints'
WIDTH = 'width'
WORKER = 'worker'
YEAR = 'year'
Y = 'y'


def mm(name): return '--' + name
def m(name): return '-' + name
def no(name): return 'no-%s' % name


VARIABLE = compile(r'(.*(?:[^$]|^))\${(\w+)\}(.*)')
MEMORY = ':memory:'


class NamespaceWithVariables(Mapping):

    def __init__(self, ns):
        self._dict = vars(ns)

    def __getitem__(self, name):
        try:
            value = self._dict[name]
        except KeyError:
            value = self._dict[sub('-', '_', name)]
        try:
            match = VARIABLE.match(value)
            while match:
                value = match.group(1) + self[match.group(2)] + match.group(3)
                match = VARIABLE.match(value)
            return sub(r'\$\$', '$', value)
        except TypeError:
            return value

    def path(self, name, index=None, rooted=True):
        # special case sqlite3 in-memory database
        if self[name] == MEMORY: return self[name]
        path = self[name]
        if index is not None: path = path[index]
        path = expanduser(path)
        if rooted and relpath(path) and name != ROOT:
            path = join(self.path(ROOT), path)
        return realpath(normpath(path))

    def file(self, name, index=None, rooted=True):
        file = self.path(name, index=index, rooted=rooted)
        # special case sqlite3 in-memory database
        if file == MEMORY: return file
        path = dirname(file)
        if not exists(path):
            makedirs(path)
        return file

    def dir(self, name, index=None, rooted=True):
        path = self.path(name, index=index, rooted=rooted)
        if not exists(path):
            makedirs(path)
        return path

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self.__dict__)


def parser():

    parser = ArgumentParser(prog=PROGNAME)

    parser.add_argument(m(F), mm(DATABASE), action='store', default='${root}/database.sqlr', metavar='FILE',
                        help='the database file')
    parser.add_argument(mm(DEV), action='store_true', help='show stack trace on error')
    parser.add_argument(mm(LOGS), action='store', default='logs', metavar='DIR',
                        help='the directory for logs')
    parser.add_argument(m(L), mm(LOG), action='store', metavar='FILE',
                        help='the file for the log (command name by default)')
    parser.add_argument(mm(ROOT), action='store', default='~/.ch2', metavar='DIR',
                        help='the root directory for the default configuration')
    parser.add_argument(m(V), mm(VERBOSITY), action='store', nargs=1, default=None, type=int, metavar='VERBOSITY',
                        help='output level for stderr (0: silent; 5:noisy)')
    parser.add_argument(m(V.upper()), mm(VERSION), action='version', version=CH2_VERSION,
                        help='display version and exit')

    subparsers = parser.add_subparsers(title='commands', dest=COMMAND)

    activities = subparsers.add_parser(ACTIVITIES, help='read activity data')
    activities.add_argument(mm(FORCE), action='store_true', help='re-read file and delete existing data')
    activities.add_argument(mm(FAST), action='store_true', help='do not calculate statistics')
    activities.add_argument(PATH, action='store', metavar='PATH', nargs='+',
                            help='path to fit file(s)')
    activities.add_argument(m(D.upper()), mm(CONSTANT), action='store', nargs='*', metavar='NAME=VALUE', dest=CONSTANT,
                            help='constant(s) to be stored with the activities')
    activities.add_argument(m(K.upper()), mm(KARG), action='store', nargs='*', metavar='NAME=VALUE', dest=KARG,
                            help='keyword argument(s) to be passed to the pipelines')
    activities.add_argument(mm(WORKER), action='store', metavar='ID', type=int,
                            help='internal use only (identifies sub-process workers)')

    constant = subparsers.add_parser(CONSTANTS, help='set and examine constants')
    constant_flags = constant.add_mutually_exclusive_group()
    constant_flags.add_argument(mm(DELETE), action='store_true', help='delete existing value(s)')
    constant_flags.add_argument(mm(SET), action='store_true', help='store a new value')
    constant.add_argument(mm(FORCE), action='store_true', help='confirm deletion(s) without value')
    constant.add_argument(NAME, action='store', nargs='?', metavar='NAME', help='constant name')
    constant.add_argument(DATE, action='store', nargs='?', metavar='DATE', help='date when measured')
    constant.add_argument(VALUE, action='store', nargs='?', metavar='VALUE', help='constant value')

    dump = subparsers.add_parser(DUMP)  # todo - this one needs tests!
    dump_format = dump.add_mutually_exclusive_group()
    dump_format.add_argument(mm(PRINT), action='store_const', dest=FORMAT, const=PRINT, help='default format')
    dump_format.add_argument(mm(CSV), action='store_const', dest=FORMAT, const=CSV, help='CVS format')
    dump_format.add_argument(mm(DESCRIBE), action='store_const', dest=FORMAT, const=DESCRIBE, help='summary format')
    dump.add_argument(mm(MAX_COLUMNS), action='store', metavar='N', type=int, help='pandas max_columns attribute')
    dump.add_argument(mm(MAX_COLWIDTH), action='store', metavar='N', type=int, help='pandas max_colwidth attribute')
    dump.add_argument(mm(MAX_ROWS), action='store', metavar='N', type=int, help='pandas max_rows attribute')
    dump.add_argument(mm(WIDTH), action='store', metavar='N', type=int, help='pandas width attribute')
    dump_sub = dump.add_subparsers(dest=SUB_COMMAND)
    dump_statistics = dump_sub.add_parser(STATISTICS)
    dump_statistics.add_argument(NAMES, action='store', nargs='*', metavar='NAME', help='statistic names')
    dump_statistics.add_argument(mm(START), action='store', metavar='TIME', help='start time')
    dump_statistics.add_argument(mm(FINISH), action='store', metavar='TIME', help='finish time')
    dump_statistics.add_argument(mm(OWNER), action='store', metavar='OWNER',
                                 help='typically the class that created the data')
    dump_statistics.add_argument(mm(CONSTRAINT), action='store', metavar='CONSTRAINT',
                                 help='a value that makes the name unique (eg activity group)')
    dump_statistics.add_argument(mm(SCHEDULE), action='store', metavar='SCHEDULE',
                                 help='the schedule on which some statistics are calculated')
    dump_statistics.add_argument(mm(SOURCE_IDS), action='store', nargs='*', metavar='ID', type=int,
                                 help='the source IDs for the statistic')
    dump_statistic_quartiles = dump_sub.add_parser(STATISTIC_QUARTILES)
    dump_statistic_quartiles.add_argument(NAMES, action='store', nargs='*', metavar='NAME', help='statistic names')
    dump_statistic_quartiles.add_argument(mm(START), action='store', metavar='TIME', help='start time')
    dump_statistic_quartiles.add_argument(mm(FINISH), action='store', metavar='TIME', help='finish time')
    dump_statistic_quartiles.add_argument(mm(OWNER), action='store', metavar='OWNER',
                                          help='typically the class that created the data')
    dump_statistic_quartiles.add_argument(mm(CONSTRAINT), action='store', metavar='CONSTRAINT',
                                          help='a value that makes the name unique (eg activity group)')
    dump_statistic_quartiles.add_argument(mm(SCHEDULE), action='store', metavar='SCHEDULE',
                                          help='the schedule on which some statistics are calculated')
    dump_statistic_quartiles.add_argument(mm(SOURCE_IDS), action='store', nargs='*', metavar='ID', type=int,
                                          help='the source IDs for the statistic')
    sump_table = dump_sub.add_parser(TABLE)
    sump_table.add_argument(NAME, action='store', metavar='NAME', help='table name')
    dump.set_defaults(format=PRINT)

    default_config = subparsers.add_parser(DEFAULT_CONFIG,
                                           help='configure the default database ' +
                                                '(see docs for full configuration instructions)')
    default_config.add_argument(mm(no(DIARY)), action='store_true', help='skip diary creation (for migration)')

    diary = subparsers.add_parser(DIARY, help='daily diary and summary')
    diary.add_argument(DATE, action='store', metavar='DATE', nargs='?',
                       help='an optional date to display (default is today)')
    diary.add_argument(mm(FAST), action='store_true',
                       help='skip update of statistics on exit')
    diary_summary = diary.add_mutually_exclusive_group()
    diary_summary.add_argument(m(M), mm(MONTH), action='store_const', dest=SCHEDULE, const='m',
                               help='show monthly summary')
    diary_summary.add_argument(m(Y), mm(YEAR), action='store_const', dest=SCHEDULE, const='y',
                               help='show yearly summary')
    diary_summary.add_argument(mm(SCHEDULE), metavar='SCHEDULE',
                               help='show summary for given schedule')

    fit = subparsers.add_parser(FIT, help='display contents of fit file')
    fit.add_argument(PATH, action='store', metavar='PATH', nargs='+',
                     help='path to fit file')
    fit_format = fit.add_argument_group(title='output format (one required)').add_mutually_exclusive_group(required=True)
    fit_format.add_argument(mm(GREP), action='store', dest=GREP, nargs='+', metavar='MSG:FLD[=VAL]',
                            help='show matching entries')
    fit_format.add_argument(mm(RECORDS), action='store_const', dest=FORMAT, const=RECORDS,
                            help='show high-level structure (ordered by time)')
    fit_format.add_argument(mm(TABLES), action='store_const', dest=FORMAT, const=TABLES,
                            help='show high-level structure (grouped in tables)')
    fit_format.add_argument(mm(CSV), action='store_const', dest=FORMAT, const=CSV,
                            help='show high-level structure (in CSV format)')
    fit_format.add_argument(mm(TOKENS), action='store_const', dest=FORMAT, const=TOKENS,
                            help='show low-level tokens')
    fit_format.add_argument(mm(FIELDS), action='store_const', dest=FORMAT, const=FIELDS,
                            help='show low-level fields (within tokens)')
    fit.add_argument(mm(AFTER_RECORDS), action='store', type=int, metavar='N', default=None,
                     help='skip initial records')
    fit.add_argument(mm(LIMIT_RECORDS), action='store', type=int, metavar='N', default=-1,
                     help='limit number of records displayed')
    fit.add_argument(mm(AFTER_BYTES), action='store', type=int, metavar='N', default=None,
                     help='skip initial bytes')
    fit.add_argument(mm(LIMIT_BYTES), action='store', type=int, metavar='N', default=-1,
                     help='limit number of bytes displayed')
    fit.add_argument(mm(INTERNAL), action='store_true',
                     help='display internal messages')
    fit.add_argument(mm(ALL_MESSAGES), action='store_true',
                     help='display undocumented messages')
    fit.add_argument(mm(ALL_FIELDS), action='store_true',
                     help='display undocumented fields')
    fit.add_argument(m(M), mm(MESSAGE), action='store', nargs='+', metavar='MSG',
                     help='display named messages (--grep, --records, --tables)')
    fit.add_argument(m(F), mm(FIELD), action='store', nargs='+', metavar='FLD',
                     help='display named fields (--grep, --records, --tables')
    fit.add_argument(m(W), mm(WARN), action='store_true',
                     help='log additional warnings')
    fit.add_argument(mm(WIDTH), action='store', type=int,
                     help='display width for some formats')
    fit.add_argument(mm(no(VALIDATE)), action='store_true',
                     help='do not validate checksum, length')
    fit.add_argument(mm(MAX_DELTA_T), action='store', type=float, metavar='S',
                     help='max seconds between timestamps (and non-decreasing)')
    fit.add_argument(mm(NAME), action='store_true',
                     help='print file name')
    fit.add_argument(mm(NOT), action='store_true',
                     help='print file names that don\'t match (--grep --name)')
    fit.add_argument(mm(MATCH), action='store', type=int, default=-1,
                     help='max number of matches (--grep, default -1 for all)')
    fit.add_argument(mm(COMPACT), action='store_true',
                     help='no space between records (--grep)')
    fit.add_argument(mm(CONTEXT), action='store_true',
                     help='display entire record (--grep)')
    fit.set_defaults(format=GREP)   # because that's the only one not set if the option is used

    fix_fit = subparsers.add_parser(FIX_FIT, help='fix a corrupted fit file')
    fix_fit.add_argument(PATH, action='store', metavar='PATH', nargs='+',
                         help='path to fit file')
    fix_fit.add_argument(m(W), mm(WARN), action='store_true',
                         help='additional warning messages')
    fix_fit_output = fix_fit.add_argument_group(title='output (default hex to stdout)').add_mutually_exclusive_group()
    fix_fit_output.add_argument(m(O), mm(OUTPUT), action='store',
                                help='output file for fixed data (otherwise, stdout)')
    fix_fit_output.add_argument(mm(DISCARD), action='store_true',
                                help='discard output (otherwise, stdout)')
    fix_fit_output.add_argument(mm(RAW), action='store_true',
                                help='raw binary to stdout (otherwise, hex encoded)')
    fix_fit_output.add_argument(mm(NAME_BAD), action='store_false', dest=NAME, default=None,
                                help='print file name if bad')
    fix_fit_output.add_argument(mm(NAME_GOOD), action='store_true', dest=NAME, default=None,
                                help='print file name if good')
    fix_fit_process = fix_fit.add_argument_group(title='processing (default disabled)')
    fix_fit_process.add_argument(mm(ADD_HEADER), action='store_true',
                                 help='preprend a new header')
    fix_fit_stage = fix_fit_process.add_mutually_exclusive_group()
    fix_fit_stage.add_argument(mm(DROP), action='store_true',
                               help='search for data that can be dropped to give a successful parse')
    fix_fit_stage.add_argument(mm(SLICES), action='store', metavar='A:B,C:D,...',
                               help='data slices to pick')
    fix_fit_stage.add_argument(mm(START), action='store', type=to_time, metavar='TIME',
                               help='change start time')
    fix_fit_process.add_argument(mm(FIX_HEADER), action='store_true',
                                 help='modify the header')
    fix_fit_process.add_argument(mm(FIX_CHECKSUM), action='store_true',
                                 help='modify the checksum')
    fix_fit_process.add_argument(mm(no(FORCE)), action='store_false', dest=FORCE,
                                 help='don\'t parse record contents')
    fix_fit_process.add_argument(mm(no(VALIDATE)), action='store_false', dest=VALIDATE,
                                 help='don\'t validate the final data')
    fix_fit_params = fix_fit.add_argument_group(title='parameters')
    fix_fit_params.add_argument(mm(HEADER_SIZE), action='store', type=int, metavar='N',
                                help='header size (validation and/or new header)')
    fix_fit_params.add_argument(mm(PROTOCOL_VERSION), action='store', type=int, metavar='N',
                                help='protocol version (validation and/or new header)')
    fix_fit_params.add_argument(mm(PROFILE_VERSION), action='store', type=int, metavar='N',
                                help='profile version (validation and/or new header)')
    fix_fit_params.add_argument(mm(MIN_SYNC_CNT), action='store', type=int, metavar='N', default=3,
                                help='minimum number of records to read for synchronization')
    fix_fit_params.add_argument(mm(MAX_RECORD_LEN), action='store', type=int, metavar='N', default=None,
                                help='maximum record length')
    fix_fit_params.add_argument(mm(MAX_DROP_CNT), action='store', type=int, metavar='N', default=1,
                                help='maximum number of gaps to drop')
    fix_fit_params.add_argument(mm(MAX_BACK_CNT), action='store', type=int, metavar='N', default=3,
                                help='maximum number of readable records to discard in a single gap')
    fix_fit_params.add_argument(mm(MAX_FWD_LEN), action='store', type=int, metavar='N', default=200,
                                help='maximum number of bytes to drop in a single gap')
    fix_fit_params.add_argument(mm(MAX_DELTA_T), action='store', type=float, metavar='S',
                                help='max number of seconds between timestamps')

    garmin = subparsers.add_parser(GARMIN, help='download monitor data from garmin connect')
    garmin.add_argument(DIR, action='store', metavar='DIR',
                        help='the directory where FIT files are stored')
    garmin.add_argument(mm(USER), action='store', metavar='USER', required=True,
                        help='garmin connect username')
    garmin.add_argument(mm(PASS), action='store', metavar='PASSWORD', required=True,
                        help='garmin connect password')
    garmin.add_argument(mm(DATE), action='store', metavar='DATE', type=to_date,
                        help='date to download')

    help = subparsers.add_parser(HELP, help='display help')
    help.add_argument(TOPIC, action='store', nargs='?', metavar=TOPIC,
                      help='the subject for help')

    jupyter = subparsers.add_parser(JUPYTER, help='access jupyter')
    jupyter_cmds = jupyter.add_subparsers(title='sub-commands', dest=SUB_COMMAND, required=True)
    jupyter_cmds.add_parser(LIST, help='list available templates')
    jupyter_show = jupyter_cmds.add_parser(SHOW, help='display a template (starting server if needed)')
    jupyter_show.add_argument(NAME, action='store', help='the template name')
    jupyter_show.add_argument(ARG, action='store', nargs='*', help='template arguments')
    jupyter_cmds.add_parser(START, help='start a background service')
    jupyter_cmds.add_parser(STOP, help='stop the background service')
    jupyter_cmds.add_parser(STATUS, help='display status of background service')
    jupyter_cmds.add_parser(SERVICE, help='internal use only - use start/stop')

    monitor = subparsers.add_parser(MONITOR, help='read monitor data')
    monitor.add_argument(mm(FORCE), action='store_true', help='re-read file and delete existing data')
    monitor.add_argument(mm(FAST), action='store_true', help='do not calculate statistics')
    monitor.add_argument(PATH, action='store', metavar='PATH', nargs='+',
                         help='path to fit file(s)')
    monitor.add_argument(m(K.upper()), mm(KARG), action='store', nargs='*', metavar='NAME=VALUE', dest=KARG,
                         help='keyword argument(s) to be passed to the pipelines')
    monitor.add_argument(mm(WORKER), action='store', metavar='ID', type=int,
                         help='internal use only (identifies sub-process workers)')

    statistics = subparsers.add_parser(STATISTICS, help='(re-)generate statistics')
    statistics.add_argument(mm(FORCE), action='store_true',
                            help='delete existing statistics')
    statistics.add_argument(mm(LIKE), action='store', metavar='PATTERN',
                            help='run only matching pipeline classes')
    statistics.add_argument(START, action='store', metavar='START', nargs='?',
                            help='optional start date')
    statistics.add_argument(FINISH, action='store', metavar='FINISH', nargs='?',
                            help='optional finish date (if start also given)')
    statistics.add_argument(m(K.upper()), mm(KARG), action='store', nargs='*', metavar='NAME=VALUE', dest=KARG,
                            help='keyword argument(s) to be passed to the pipelines')
    statistics.add_argument(mm(WORKER), action='store', metavar='ID', type=int,
                            help='internal use only (identifies sub-process workers)')

    noop = subparsers.add_parser(NO_OP,
                                 help='used within jupyter (no-op from cmd line)')

    package_fit_profile = subparsers.add_parser(PACKAGE_FIT_PROFILE,
                                                help='parse and save the global fit profile (dev only)')
    package_fit_profile.add_argument(PATH, action='store', metavar='PROFILE',
                                     help='the path to the profile (Profile.xlsx)')
    package_fit_profile.add_argument(m(W), mm(WARN), action='store_true',
                                     help='additional warning messages')

    test_schedule = subparsers.add_parser(TEST_SCHEDULE, help='print schedule locations in a calendar')
    test_schedule.add_argument(SCHEDULE, action='store', metavar='SCHEDULE',
                               help='the schedule to test')
    test_schedule.add_argument(mm(START), action='store', metavar='DATE',
                               help='the date to start displaying data')
    test_schedule.add_argument(mm(MONTHS), action='store', metavar='N', type=int,
                               help='the number of months to display')

    unlock = subparsers.add_parser(UNLOCK, help='remove database locking')

    return parser


def bootstrap_file(file, *args, configurator=None, post_config=None):

    from ..lib.log import make_log
    from ..config.database import config
    from ..squeal.database import Database

    args = [mm(DATABASE), file.name] + list(args)
    if configurator:
        db = config(*args)
        configurator(db)
    args += post_config if post_config else []
    args = NamespaceWithVariables(parser().parse_args(args))
    make_log(args)
    db = Database(args, log)

    return args, db


def parse_pairs(pairs):
    # simple name, value pairs.  owner and constraint supplied by command.
    d = {}
    if pairs is not None:
        for pair in pairs:
            name, value = pair.split('=', 1)
            for type in (int, float, to_time):
                try:
                    value = type(value)
                    break
                except ValueError:
                    pass
            d[name] = value
    return d




from logging import basicConfig, getLogger, INFO, DEBUG
from os.path import join
from sys import stdout
from unittest import TestCase

from ch2.commands.args import RECORDS, TABLES
from ch2.fit.fix import fix
from ch2.fit.format.tokens import FileHeader
from ch2.fit.profile.profile import read_fit
from ch2.fit.summary import summarize
from ch2.lib.tests import OutputMixin


class TestFixFit(TestCase, OutputMixin):

    def setUp(self):
        if not getLogger().handlers:
            basicConfig(stream=stdout, level=DEBUG)
        self.log = getLogger()
        self.test_dir = 'data/test'
        self.profile_path = 'data/sdk/Profile.xlsx'

    def test_null(self):
        good = read_fit(self.log, join(self.test_dir, 'source/personal/2018-08-27-rec.fit'))
        same = fix(bytearray(good))
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(good, same)

    def test_null_drop(self):
        good = read_fit(self.log, join(self.test_dir, 'source/personal/2018-08-27-rec.fit'))
        same = fix(bytearray(good), drop=True, fix_checksum=True, fix_header=True)
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(good, same)

    def test_null_slices(self):
        good = read_fit(self.log, join(self.test_dir, 'source/personal/2018-08-27-rec.fit'))
        same = fix(bytearray(good), slices=':', fix_checksum=True, fix_header=True)
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(good, same)

    def test_drop(self):
        bad = read_fit(self.log, join(self.test_dir, 'source/other/8CS90646.FIT'))
        fixed = fix(bad, drop=True, fix_checksum=True, fix_header=True)
        self.assertTrue(len(fixed) < len(bad))
        with self.assertTextMatch('data/test/target/other/TestFixFit.test_drop') as output:
            summarize(RECORDS, fixed, output=output)

    def test_slices(self):
        bad = read_fit(self.log, join(self.test_dir, 'source/other/8CS90646.FIT'))
        with self.assertRaisesRegex(Exception, 'Error fixing checksum'):
            fix(bad, slices=':1000', fix_checksum=True, fix_header=True)  # first 1k bytes only

    def test_no_last_byte(self):
        good = read_fit(self.log, join(self.test_dir, 'source/personal/2018-08-27-rec.fit'))
        same = fix(bytearray(good), drop=True, fix_checksum=True, fix_header=True)
        self.assertEqual(same, good)
        fixed = fix(bytearray(good)[:-1], drop=True, fix_checksum=True, fix_header=True)
        self.assertEqual(fixed, good)
        fixed = fix(bytearray(good)[:-2], drop=True, fix_checksum=True, fix_header=True)
        self.assertEqual(fixed, good)

    def test_no_header(self):
        good = read_fit(self.log, join(self.test_dir, 'source/personal/2018-08-27-rec.fit'))
        same = fix(bytearray(good), drop=True, fix_checksum=True)
        self.assertTrue(good is not same)  # check making a copy
        self.assertEqual(same, good)
        header = FileHeader(good)
        with self.assertRaisesRegex(Exception, 'Error fixing checksum'):
            fix(bytearray(good)[len(header):], fix_checksum=True, fix_header=True)
        fixed = fix(bytearray(good)[len(header):],
                    add_header=True, drop=True, fix_checksum=True, fix_header=True)
        self.assertEqual(good, fixed)
        fixed = fix(bytearray(good), add_header=True, slices=':14,28:', fix_checksum=True, fix_header=True)
        self.assertEqual(good, fixed)

    def test_other_header(self):
        bad = read_fit(self.log, join(self.test_dir, 'source/other/8CS90646.FIT'))
        old_header = FileHeader(bad)
        fixed = fix(bytearray(bad), drop=True, header_size=27, fix_checksum=True, fix_header=True)
        new_header = FileHeader(fixed)
        self.assertEqual(new_header.header_size, 27)
        self.assertEqual(new_header.protocol_version, old_header.protocol_version)
        self.assertEqual(new_header.profile_version, old_header.profile_version)

    def test_other_bad(self):
        bad = read_fit(self.log, join(self.test_dir, 'source/other/2018-04-15-09-18-20.fit'))
        fixed = fix(bytearray(bad), drop=True, fix_checksum=True, fix_header=True)
        with self.assertTextMatch(join(self.test_dir, 'target/other/TestFixFit.test_unknown_bad:1')) as output:
            summarize(RECORDS, fixed, output=output)
        bad = read_fit(self.log, join(self.test_dir, 'source/other/2018-02-24-10-04-10.fit'))
        fixed = fix(bytearray(bad), drop=True, fix_checksum=True, fix_header=True)
        with self.assertTextMatch(join(self.test_dir, 'target/other/TestFixFit.test_unknown_bad:2')) as output:
            summarize(RECORDS, fixed, output=output)

    def test_other_good(self):
        good = read_fit(self.log, join(self.test_dir, 'source/other/77F73023.FIT'))
        same = fix(bytearray(good))
        self.assertEqual(good, same)
        self.assertFalse(good is same)

    def test_pyfitparse_fix_header(self):
        for file in ('activity-filecrc.fit',  # bad checksum
                     'activity-activity-filecrc.fit',  # data size incorrect
                     'activity-settings.fit',  # data size incorrect
                     ):
            bad = read_fit(self.log, join(self.test_dir, 'source/python-fitparse', file))
            with self.assertBinaryMatch(join(self.test_dir, 'source/python-fitparse-fix', file)) as output:
                output.write(fix(bad, fix_checksum=True, fix_header=True))

    def test_pyfitparse_fix_drop(self):
        for file in ('activity-unexpected-eof.fit',  # data size incorrect
                     'activity-settings-nodata.fit',   # data size incorrect
                     'activity-settings-corruptheader.fit',  # data size incorrect
                     ):
            bad = read_fit(self.log, join(self.test_dir, 'source/python-fitparse', file))
            with self.assertBinaryMatch(join(self.test_dir, 'source/python-fitparse-fix', file)) as output:
                output.write(fix(bad, drop=True, fix_checksum=True, fix_header=True))

    def test_pyfitparse_fix_drop_2(self):
        for file in ('event_timestamp.fit',  # data size incorrect
                     'antfs-dump.63.fit',  # strange timestamp
                     'compressed-speed-distance.fit',  # strange timestamp
                     ):
            bad = read_fit(self.log, join(self.test_dir, 'source/python-fitparse', file))
            with self.assertBinaryMatch(join(self.test_dir, 'source/python-fitparse-fix', file)) as output:
                output.write(fix(bad, drop=True, max_drop_cnt=2, fix_checksum=True, fix_header=True))

    def test_drop_bug(self):
        bad = read_fit(self.log, join(self.test_dir, 'source/personal/2018-07-26-rec.fit'))
        fix(bad, drop=True, fix_checksum=True, fix_header=True, max_delta_t=60, max_fwd_len=500)




from abc import abstractmethod
from collections import defaultdict, Counter
from logging import getLogger
from re import sub
from struct import unpack, pack

from .records import LazyRecord, merge_duplicates
from ..profile.fields import TypedField, TIMESTAMP_GLOBAL_TYPE, DynamicField, CompositeField
from ..profile.types import timestamp_to_time, time_to_timestamp
from ...lib.data import WarnDict, tohex
from ...stoats.names import S

log = getLogger(__name__)
FIELD_DESCRIPTION = 206


class Identity:
    '''
    A unique ID that identifies a particular message type (a combination of message name
    and the particular definition used).
    '''

    def __init__(self, name, counter):
        self.name = name
        counter[name] += 1
        self.count = counter[name]
        self.__counter = counter

    def __repr__(self):
        current = self.__counter[self.name]
        if current == 1:
            return self.name
        else:
            return '%s (defn %d/%d)' % (self.name, self.count, current)


class Token:
    '''
    A chunk of the file.  Subclasses into different token types below.

    tag - a simple string describing the type of token (HDR etc)
    is_user - does this contain user data? (alternatively, it;s internal data for parsing)
    data - the bytes from the input file.
    '''

    __slots__ = ('tag', 'is_user', 'data')

    def __init__(self, tag, is_user, data):
        self.tag = tag
        self.is_user = is_user
        self.data = data

    def __str__(self):
        return '%s %s' % (self.tag, tohex(self.data))

    def __len__(self):
        return len(self.data)

    def _fake_field(self, key, value):
        if isinstance(value, tuple) or isinstance(value, list):
            values, units = value
            if not isinstance(values, tuple) and not isinstance(values, list):
                values = (values,)
        else:
            values, units = (value,), None
        return key, (values, units)

    def _fake_record(self, name, **kargs):
        return LazyRecord(name, None, Identity(name, Counter()), None,
                          [self._fake_field(key, value) for key, value in kargs.items()])

    @abstractmethod
    def parse_token(self, **options):
        raise NotImplementedError()

    def _describe_csv_header(self, record):
        yield self.__class__.__name__

    def _describe_csv_data(self, record, field_names=None):
        for name, (values, units) in record.data.items():
            if not field_names or name in field_names:
                yield name
                yield '' if values is None else '|'.join(str(value) for value in values)
                yield '' if units is None else units

    def describe_csv(self, record_names=None, field_names=None, internal=False, **options):
        record = self.parse_token(map_values=False, raw_time=True, rtn_composite=True,
                                  **options).force(merge_duplicates)
        if not record_names or record.name in record_names:
            yield from self._describe_csv_header(record)
            yield from self._describe_csv_data(record, field_names=field_names)

    def describe_fields(self, types):
        for name, (values, units) in self.parse_token(raw_data=True).data:
            value = tohex(values[0])
            yield '%s - %s' % (value, sub('_', ' ', name))

    def repair_timestamp(self, timestamp):
        pass


class ValidateToken(Token):
    '''
    A Token that can be validated.  Provides support for quiet or noisy validation.
    '''

    @abstractmethod
    def validate(self, data, log, quiet=False):
        raise NotImplementedError()

    def _error(self, msg, log, quiet):
        if quiet:
            log.warning(msg)
        else:
            raise Exception(msg)


FIT = b'.FIT'


class FileHeader(ValidateToken):
    '''
    The (single) header Token that starts a file.

    Contains extra functionality to allow modification when fixing FIT files.
    '''

    def __init__(self, data):
        super().__init__('HDR', False, data[:data[0]])
        self.header_size = data[0]
        self.protocol_version = data[1]
        self.profile_version = unpack('<H', data[2:4])[0]
        self.data_size = unpack('<I', data[4:8])[0]
        self.data_type = b''.join(unpack('4c', data[8:12]))
        if len(self) > 13:
            self.checksum = unpack('<H', data[12:14])[0]
            self.has_checksum = self.checksum != 0
        else:
            self.has_checksum = False

    def validate(self, data, log, quiet=False, header_size=None, protocol_version=None, profile_version=None):
        if self.header_size < 12:
            self._error('Header size too small (%d)' % self.header_size, log, quiet)
        if self.header_size not in (12, 14):
            log.warning('Header size not standard (%d/12-14)' % self.header_size)
        if header_size is not None:
            if self.header_size != header_size:
                self._error('Header size incorrect (%d/%d)' % (self.header_size, header_size), log, quiet)
        if protocol_version is not None:
            if self.profile_version != protocol_version:
                self._error('Protocol version incorrect (%d%d)' % (self.protocol_version, protocol_version), log, quiet)
        if profile_version is not None:
            if self.profile_version != profile_version:
                self._error('Profile version incorrect (%d%d)' % (self.profile_version, profile_version), log, quiet)
        if len(data) != self.data_size + len(self) + 2:
            self._error('Data size incorrect (%d/%d+%d+2=%d)' % (len(data), self.data_size, len(self),
                                                                 self.data_size + len(self) + 2), log, quiet)
        if self.data_type != FIT:
            self._error('Data type incorrect (%s)' % (self.data_type,), log, quiet)
        if self.has_checksum:
            checksum = Checksum.crc(data[0:12])
            if checksum != self.checksum and self.checksum:
                self._error('Inconsistent checksum (%04x/%04x)' % (checksum, self.checksum), log, quiet)

    def repair(self, data, log, header_size=None, protocol_version=None, profile_version=None):
        if header_size is not None and self.header_size != header_size:
            log.warning('Changing header size: %d -> %d' % (self.header_size, header_size))
            self.header_size = header_size
            self.data = [0] * self.header_size
            self.data[0] = self.header_size
        # next two are applied even if unchanged because resizing removes
        if protocol_version is not None:
            if protocol_version != self.protocol_version:
                log.warning('Changing protocol version: %d -> %d' % (self.protocol_version, protocol_version))
            self.protocol_version = protocol_version
            self.data[1] = self.protocol_version
        if profile_version is not None:
            if profile_version != self.profile_version:
                log.warning('Changing profile version: %d -> %d' % (self.profile_version, profile_version))
            self.profile_version = profile_version
            self.data[2:4] = pack('<H', self.profile_version)
        data_size = len(data) - len(self) - 2
        if data_size != self.data_size:
            log.warning('Fixing header data size: %d -> %d' % (self.data_size, data_size))
            self.data_size = data_size
            self.data[4:8] = pack('<I', self.data_size)
        if self.data_type != FIT:
            log.warning('Fixing header data type: %s -> %s' % (self.data_type, FIT))
            self.data_type = FIT
            self.data[8:12] = self.data_type
        if self.header_size >= 14:
            checksum = Checksum.crc(self.data[0:12])
            if not self.has_checksum or checksum != self.checksum:
                log.warning('Fixing header checksum: %04x -> %04x' %
                            (self.checksum if self.has_checksum else 0, checksum))
                self.checksum = checksum
                self.data[12:14] = pack('<H', checksum)
                self.has_checksum = True

    def parse_token(self, raw_data=False, **options):
        data = {'header_size': self.data[0:1] if raw_data else self.header_size,
                'protocol_version': self.data[1:2] if raw_data else self.protocol_version,
                'profile_version': self.data[2:4] if raw_data else self.profile_version,
                'data_size': (self.data[4:8] if raw_data else self.data_size, 'bytes'),
                'data_type': self.data[8:12] if raw_data else self.data_type}
        if self.has_checksum:
            data['checksum'] = self.data[12:14] if raw_data else self.checksum
        return self._fake_record('file_header', **data)


class Defined(Token):
    '''
    A chunk of data from a FIT file that is associated with a single local message.  Bundled with
    the Definition for that type, which itself contains a reference to the Message and Fields the data contain.
    '''

    __slots__ = ('definition', 'timestamp', '_accumulators')

    def __init__(self, tag, data, state, local_message_type):
        self.definition = state.definitions[local_message_type]
        if self.definition.timestamp_field:
            self.__parse_timestamp(data, state)
        self.timestamp = state.timestamp
        self._accumulators = state.accumulators
        if len(data) < self.definition.size:
            raise Exception('Insufficient data for %s (%d/%d)' %
                            (self.definition.identity, len(data), self.definition.size))
        super().__init__(tag, True, data[0:self.definition.size])

    def __parse_timestamp(self, data, state):
        field = self.definition.timestamp_field
        times = field.field.type.parse_type(data[field.start:field.finish], 1, self.definition.endian, state.timestamp,
                                            check_bad=False)
        if times:
            state.timestamp = times[0]
        else:
            raise Exception('Could not parse timestamp')

    def parse_token(self, **options):
        return self.definition.message.parse_message(self.data, self.definition, self.timestamp,
                                                     accumulators=self._accumulators, **options)

    def describe_fields(self, types):
        yield '%s - header (local message type %d - %s)' % \
              (tohex(self.data[0:1]), self.data[0] & 0x0f, self.definition.identity)
        for field in sorted(self.definition.fields, key=lambda field: field.start):
            if field.name == 'timestamp':
                yield '%s - %s (%s) %s' % (tohex(self.data[field.start:field.finish]), field.name,
                                           field.base_type.name, self.timestamp)
            else:
                yield '%s - %s (%s)' % (tohex(self.data[field.start:field.finish]), field.name, field.base_type.name)

    def _describe_csv_header(self, record):
        yield self.__class__.__name__
        yield self.definition.local_message_type
        yield record.name

    def repair_timestamp(self, state):
        if self.definition.timestamp_field:
            field = self.definition.timestamp_field
            self.data[field.start:field.finish] = field.field.type.pack_type((state.timestamp,),
                                                                             field.count, self.definition.endian)


class DeveloperField(Defined):

    def __init__(self, data, state):
        super().__init__('FLD', data, state, data[0] & 0x0f)
        self.__parse_field_definition(state)
        self.is_user = False

    def __parse_field_definition(self, state):
        record = self.parse_token().force()
        developer_index = record.attr.developer_data_index[0][0]
        number = record.attr.field_definition_number[0][0]
        base_type = state.types.profile_to_type(record.attr.fit_base_type_id[0][0])
        # todo - more fields (optional)
        name = record.attr.field_name[0][0]
        units = record.attr.units[0][0]
        state.dev_fields[developer_index][number] = \
            TypedField(log, name, number, units, None, None, None, base_type.name, state.types)

    # todo - display differently?


class Data(Defined):

    __slots__ = ()

    def __init__(self, data, state):
        super().__init__('DTA', data, state, data[0] & 0x0f)


class CompressedTimestamp(Defined):

    __slots__ = ()

    def __init__(self, data, state):
        offset = data[0] & 0x1f
        if not state.timestamp:
            raise Exception('Compressed timestamp with no preceding absolute timestamp')
        timestamp = time_to_timestamp(state.timestamp)
        rollover = offset < timestamp & 0x1f
        state.timestamp = timestamp_to_time((timestamp & 0xffffffe0) + offset + (0x20 if rollover else 0))
        super().__init__('DTT', data, state, (data[0] & 0x60) >> 5)

    def parse_token(self, raw_time=False, **options):
        timestamp = time_to_timestamp(self.timestamp) if raw_time else self.timestamp
        if self.definition.timestamp_field:
            extra = {}
        else:
            extra = {'timestamp': ((timestamp,), S)}
        return self.definition.message.parse_message(self.data, self.definition, self.timestamp,
                                                     extra=extra, raw_time=raw_time,
                                                     accumulators=self._accumulators, **options)

    def describe_fields(self, types):
        yield '%s - header (local message type %d - %s; time delta %d)' % \
              (tohex(self.data[0:1]), (self.data[0] & 0x60) >> 5, self.definition.message.name, self.data[0] & 0x1f)
        for field in sorted(self.definition.fields, key=lambda field: field.start):
            if field.name == 'timestamp':
                yield '%s - %s (%s) %s' % (tohex(self.data[field.start:field.finish]), field.name,
                                           field.base_type.name, self.timestamp)
            else:
                yield '%s - %s (%s)' % (tohex(self.data[field.start:field.finish]), field.name, field.base_type.name)

    def repair_timestamp(self, state):
        timestamp = time_to_timestamp(state.timestamp)
        self.data[0] = (self.data[0] & 0xe0) | (timestamp & 0x1f)
        super().repair_timestamp(state)  # any fields too


class Field:

    def __init__(self, size, field, base_type):

        self.size = size
        self.field = field
        self.name = self.field.name if self.field else 'unknown'
        self.base_type = base_type

        self.count = self.size // base_type.n_bytes
        # set by definition later
        self.start = 0
        self.finish = 0


class Definition(Token):

    '''
    Encapsulate the information from the FIT file that selects as Message and a subset of Fields and
    associates them with a particular local_message_type.

    This definition is stored in the state so that when that message type is found it can be used to
    parse the data.
    '''

    def __init__(self, data, state, overhead=6, tag='DFN'):
        self.local_message_type = data[0] & 0x0f
        self.is_user = False
        self.references = set()
        self.timestamp_field = None
        self.endian = data[2] & 0x01
        self.global_message_no = unpack('<>'[self.endian]+'H', data[3:5])[0]
        self.message = state.messages.number_to_message(self.global_message_no)
        self.identity = Identity(self.message.name, state.definition_counter)
        self.fields = self.__process_fields(self._make_fields(data, state), state)
        self.accumulators = state.accumulators
        super().__init__(tag, False, data[0:overhead+3*len(self.fields)])
        state.definitions[self.local_message_type] = self

    def _make_fields(self, data, state):
        yield from self.__fields(data, state.types)

    def __fields(self, data, types):
        for i in range(data[5]):
            yield self.__field(data[6 + i * 3:6 + (i + 1) * 3], self.message, types)

    def __field(self, data, message, types):
        number, size, base = data
        try:
            field = message.number_to_field(number)
        except KeyError:
            field = None
        base_type = types.base_types[base & 0xf]
        return Field(size, field, base_type)

    def __process_fields(self, fields, state):
        offset = 1  # header
        fields = tuple(fields)
        for field in fields:
            field.start = offset
            offset += field.size
            field.finish = offset
            if field.field:  # when is this not true?
                field.field.register_accumulator(state.accumulators)
                if field.field.number == TIMESTAMP_GLOBAL_TYPE or field.field.name == 'timestamp_16':
                    self.timestamp_field = field
                if isinstance(field.field, DynamicField):
                    self.references.update(field.field.references)
        self.size = offset
        return tuple(self.__sorted(fields))

    def __provided_by(self, field):
        yield field.name
        if isinstance(field.field, CompositeField):
            for _, component in field.field._components:
                yield component.name

    def __sorted(self, fields):

        by_name = dict((field.name, field) for field in fields)
        names = list(field.name for field in fields)
        providers = defaultdict(list)
        for field in fields:
            for provided in self.__provided_by(field):
                providers[provided].append(field.name)
        references = dict((field.name, field.field.references)
                          for field in fields if isinstance(field.field, DynamicField))

        def follow(name, chain=()):
            if name in chain:
                raise Exception('Circular dependency: %s/%s' % (chain, name))
            chain = chain + (name,)
            for reference in references.get(name, []):
                for provider in providers.get(reference, []):
                    if provider in names:
                        yield from follow(provider, chain=chain)
            names.remove(name)
            yield by_name[name]

        while names:
            yield from follow(names[0])

    def describe_csv(self, record_names=None, field_names=None, internal=False, **options):
        if internal and (not record_names or self.message.name in record_names):
            yield 'Definition'
            yield self.local_message_type
            yield self.message.name
            for field in self.fields:
                if not field_names or field.name in field_names:
                    yield field.name
                    yield field.count
                    yield ''

    def describe_fields(self, types):
        # something of a hack - we pack data in odd places below just to unpack here
        for name, (values, units) in self.parse_token(raw_data=True).data:
            if len(values) > 1:
                (value, extra), padding = values, units
                extra = ': ' + extra
            else:
                value, extra, padding = values[0], '', ''
            value = tohex(value)
            yield '%s%s - %s%s' % (padding, value, sub('_', ' ', name), extra)

    def parse_token(self, raw_data=False, **options):
        data = {'local_message_type': ((self.data[0:1],
                                        str(self.local_message_type)), '') if raw_data else self.local_message_type,
                'reserved': self.data[1:2],
                'architecture': self.data[2:3],
                'message_number': ((self.data[3:5], self.message.name), '') if raw_data else self.global_message_no,
                'no_of_fields': self.data[5:6] if raw_data else self.data[5]}
        if not raw_data:
            data['message_name'] = self.message.name
        for i, field in enumerate(self.fields):
            mult = '' if field.count == 1 else 'x%d' % field.count
            desc = '%s (%s%s)' % (field.name, field.base_type.name, mult)
            raw = self.data[6 + i*3:9 + i*3]
            data['field_%d' % i] = ((raw, desc), '  ') if raw_data else desc
        return self._fake_record('definition', **data)


class DeveloperDefinition(Definition):

    def __init__(self, data, state):
        super().__init__(data, state, overhead=7, tag='DFX')

    def _make_fields(self, data, state):
        yield from super()._make_fields(data, state)
        for field_data in self.__field_data(data):
            yield self.__field(field_data, state.dev_fields)

    def __field_data(self, data):
        offset = data[5] * 3 + 7
        n_dev_fields = data[offset-1]
        for i in range(n_dev_fields):
            yield data[offset + i * 3:offset + (i + 1) * 3]

    def __field(self, data, dev_fields):
        number, size, developer_index = data
        field = dev_fields[developer_index][number]
        return Field(size, field, field.type)

    def describe_fields(self, types):
        yield from super().describe_fields(types)
        offset = self.data[5] * 3 + 7
        yield '%s - no of dev fields' % tohex(self.data[offset-1:offset])
        for field_data in self.__field_data(self.data):
            fdn, size, ddi = field_data
            yield '  %s - dev fld %d/%d' % (tohex(field_data), fdn, ddi)


class Checksum(ValidateToken):

    @staticmethod
    def crc(data):

        CRC = [0x0000, 0xCC01, 0xD801, 0x1400, 0xF001, 0x3C00, 0x2800, 0xE401,
               0xA001, 0x6C00, 0x7800, 0xB401, 0x5000, 0x9C01, 0x8801, 0x4400]

        checksum = 0
        for byte in data:
            tmp = CRC[checksum & 0xf]
            checksum = (checksum >> 4) & 0xfff
            checksum = checksum ^ tmp ^ CRC[byte & 0xf]
            tmp = CRC[checksum & 0xf]
            checksum = (checksum >> 4) & 0xfff
            checksum = checksum ^ tmp ^ CRC[(byte >> 4) & 0xf]
        return checksum

    def __init__(self, data):
        super().__init__('CRC', False, data)
        self.checksum = unpack('<H', self.data)[0]

    def validate(self, all_data, log, quiet=False):
        checksum = self.crc(all_data[:-2])
        if checksum != self.checksum:
            self._error('Bad checksum (%04x/%04x)' % (checksum, self.checksum), log, quiet)

    def repair(self, data, log):
        checksum = self.crc(data[:-2])
        if checksum != self.checksum:
            log.warning('Fixing final checksum: %04x -> %04x' % (self.checksum, checksum))
            self.checksum = checksum
            self.data[:2] = pack('<H', checksum)

    def parse_token(self, raw_data=False, **options):
        return self._fake_record('checksum', checksum=self.data[0:2] if raw_data else self.checksum)


def token_factory(data, state):
    header = data[0]
    if header & 0x80:
        return CompressedTimestamp(data, state)
    else:
        if header & 0x40:
            if header & 0x20:
                return DeveloperDefinition(data, state)
            else:
                return Definition(data, state)
        else:
            if header & 0x10:
                log.debug('Reserved bit set')
            token = Data(data, state)
            if token.definition.global_message_no == FIELD_DESCRIPTION:
                return DeveloperField(data, state)
            else:
                return token


class State:

    def __init__(self, types, messages, max_delta_t=None):
        self.types = types
        self.messages = messages
        self.max_delta_t = max_delta_t
        self.dev_fields = defaultdict(lambda: WarnDict(log, 'No definition for developer field %s'))
        self.definitions = WarnDict(log, 'No definition for local message type %s')
        self.definition_counter = Counter()
        self.accumulators = {}
        self._timestamp = None

    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp):
        self._timestamp = self._validate_timestamp(timestamp)

    def _validate_timestamp(self, timestamp):
        if self.max_delta_t and self._timestamp:
            if (timestamp - self._timestamp).total_seconds() > self.max_delta_t:
                raise Exception('Too large shift in timestamp (%.1fs: %s/%s' %
                                ((timestamp - self.timestamp).total_seconds(), self._timestamp, timestamp))
            if timestamp < self._timestamp:
                raise Exception('Timestep decreased (%s/%s)' % (self._timestamp, timestamp))
        return timestamp

    def copy(self):
        copy = State(self.types, self.messages, self.max_delta_t)
        copy.dev_fields.update(self.dev_fields)
        copy.definitions.update(self.definitions)
        copy.definition_counter.update(self.definition_counter)
        copy.accumulators.update(self.accumulators)
        copy._timestamp = self._timestamp
        return copy




from logging import getLogger

from . import ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator
from ..names import RAW_ELEVATION, ELEVATION, DISTANCE, M
from ...data.elevation import fix_elevation
from ...data.frame import activity_statistics, present
from ...squeal import StatisticJournalFloat

log = getLogger(__name__)


class ElevationCalculator(ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, smooth=3, **kargs):
        self.smooth = smooth
        super().__init__(*args, **kargs)

    def _read_dataframe(self, s, ajournal):
        try:
            df = activity_statistics(s, DISTANCE, RAW_ELEVATION, ELEVATION,
                                     activity_journal=ajournal, with_timespan=True)
            return df
        except Exception as e:
            log.warning(f'Failed to generate statistics for elevation: {e}')
            raise

    def _calculate_stats(self, s, ajournal, df):
        return fix_elevation(df, smooth=self.smooth)

    def _copy_results(self, s, ajournal, loader, df):
        for time, row in df.iterrows():
            loader.add(ELEVATION, M, None, ajournal.activity_group, ajournal, row[ELEVATION], time,
                       StatisticJournalFloat)




import setuptools

setuptools.setup(name='choochoo',
                 packages=setuptools.find_packages(),
                 version='0.19.12',
                 author='andrew cooke',
                 author_email='andrew@acooke.org',
                 description='Data Science for Training',
                 url='https://github.com/andrewcooke/choochoo',
                 long_description='''
# Choochoo (ch2)

An **open**, **hackable** and **free** training diary.

See [documentation](https://andrewcooke.github.io/choochoo/) for full
details.

Source and screenshots on [github](https://github.com/andrewcooke/choochoo).
                 ''',
                 long_description_content_type='text/markdown',
                 include_package_data=True,
                 install_requires=[
                     'cachetools',
                     'urwid',
                     'sqlalchemy',
                     'openpyxl',
                     'numpy',
                     'pandas',
                     'pyGeoTile',
                     'colorama',
                     'pendulum',
                     'requests',
                     'scipy',
                     'bokeh',
                     'jupyter'
                     ],
                 entry_points={
                     'console_scripts': [
                         'ch2 = ch2:main',
                     ],
                 },
                 classifiers=(
                     "Programming Language :: Python :: 3.7",
                     "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                     "Operating System :: OS Independent",
                     "Development Status :: 4 - Beta",
                 ),
                 )





from .file import SRTM1_DIR, SAMPLES, ElevationSupport, elevation_from_constant


def bilinear_elevation_from_constant(log, s, dir_name=SRTM1_DIR):
    return elevation_from_constant(log, s, BilinearElevation, dir_name=dir_name)


class BilinearElevation(ElevationSupport):
    '''
    Provide elevation data from the files in `dir` which should be hgt files with standard naming,
    either zipped or unzipped, downloaded from http://dwtkns.com/srtm30m/.

    All coords are "GPS coords" afaict.

    If dir is None then None will be return as elevation for all queries.

    If dir is not None and a file is missing for a particular lat/lon then an exception is raised.

    Elevations are bilinear interpolated from the surrounding arcsec grid.
    '''

    def elevation(self, lat, lon):
        if self._dir:
            flat, flon, h = self._lookup(lat, lon)
            x = (lon - flon) * (SAMPLES - 1)
            y = (lat - flat) * (SAMPLES - 1)  # -1 because weird inclusive-at-each-side tiling
            i, j = int(x), int(y)
            # bilinear
            # it's ok to use +1 blindly here because we're never on the top/right cells or we'd be in
            # a different tile.
            k = y - j
            h0 = h[j, i] * (1-k) + h[j+1, i] * k
            h1 = h[j, i+1] * (1-k) + h[j+1, i+1] * k
            k = x - i
            return h0 * (1-k) + h1 * k
        else:
            return None




from abc import abstractmethod
from logging import getLogger
from time import time

from psutil import cpu_count

from .load import StatisticJournalLoader
from ..lib.date import format_seconds
from ..lib.utils import short_str
from ..lib.workers import Workers
from ..squeal import Pipeline
from ..squeal.types import short_cls

log = getLogger(__name__)
CPU_FRACTION = 0.9
MAX_REPEAT = 3
NONE = object()


def run_pipeline(db, type, like=None, id=None, **extra_kargs):
    with db.session_context() as s:
        for pipeline in Pipeline.all(s, type, like=like, id=id):
            kargs = dict(pipeline.kargs)
            kargs.update(extra_kargs)
            log.info(f'Running {short_cls(pipeline.cls)}({short_str(pipeline.args)}, {short_str(kargs)}')
            log.debug(f'Running {pipeline.cls}({pipeline.args}, {kargs})')
            start = time()
            pipeline.cls(db, *pipeline.args, id=pipeline.id, **kargs).run()
            duration = time() - start
            log.info(f'Ran {short_cls(pipeline.cls)} in {format_seconds(duration)}')


class BasePipeline:

    def __init__(self, *args, **kargs):
        if args or kargs:
            log.warning(f'Unused ({type(self)}): {args} {list(kargs.keys())}')

    def _assert(self, name, value):
        if value is None:
            raise Exception(f'Undefined {name}')
        else:
            return value


class MultiProcPipeline(BasePipeline):

    def __init__(self, db, *args, owner_out=None, force=False,
                 overhead=1, cost_calc=20, cost_write=1, n_cpu=None, worker=None, id=None, **kargs):
        self._db = db
        self.owner_out = owner_out or self  # the future owner of any calculated statistics
        self.force = force  # force re-processing
        self.overhead = overhead  # next three args are used to decide if workers are needed
        self.cost_calc = cost_calc  # see _cost_benefit for full details
        self.cost_write = cost_write  # defaults guarantee a single thread
        self.n_cpu = max(1, int(cpu_count() * CPU_FRACTION)) if n_cpu is None else n_cpu  # number of cpus available
        self.worker = worker  # if True, then we're in a sub-process
        self.id = id  # the id for the pipeline entry in the database (passed to sub-processes)
        super().__init__(*args, **kargs)

    def run(self):
        with self._db.session_context() as s:
            self._startup(s)

            if self.force:
                if self.worker:
                    log.warning('Worker deleting data')
                self._delete(s)

            missing = self._missing(s)
            log.debug(f'Have {len(missing)} missing ranges')

            if self.worker:
                log.debug('Worker, so execute directly')
                self._run_all(s, missing)
            elif not missing:
                log.info(f'No missing data for {short_cls(self)}')
            else:
                n_total, n_parallel = self.__cost_benefit(missing, self.n_cpu)
                if n_parallel < 2 or len(missing) == 1:
                    self._run_all(s, missing)
                else:
                    self.__spawn(s, missing, n_total, n_parallel)
            self._shutdown(s)

    def _run_all(self, s, missing):
        for missed in missing:
            log.debug(f'Run {missed}')
            self._run_one(s, missed)
            s.commit()

    def _startup(self, s):
        pass

    def _shutdown(self, s):
        pass

    # as a general rule, _missing and _args should be implemented together
    @abstractmethod
    def _missing(self, s):
        raise NotImplementedError()

    @abstractmethod
    def _delete(self, s):
        raise NotImplementedError()

    @abstractmethod
    def _run_one(self, s, missed):
        raise NotImplementedError()

    def __cost_benefit(self, missing, n_cpu):

        # is it worth using workers?  there's some cost in starting them up and there will be contention
        # in accessing the database.
        # let's say COST (for one missing time) is COST_WRITE + COST_CALC (ignoring units).
        # if we have N_MISSING tasks divided into N_TOTAL workloads amongst N_PARALLEL workers
        # then we have these conditions (in order):
        #   N_PARALLEL * (COST_WRITES/ COST) <= 1 so that we avoid blocking completely on writes
        #   N_PARALLEL <= N_CPU
        #   N_TOTAL <= N_MISSING / N_PARALLEL
        #   (N_MISSING / N_TOTAL) * COST > OVERHEAD so we're not wasting our time
        #   N_TOTAL <= N_PARALLEL * MAX_REPEAT because we want large batches, but not too large
        # really we should include estimates of disk and cpu speed here, in which case we need to separate
        # out COST_READ too (currently folded into COST_CALC).

        log.debug(f'Batching for n_cpu={n_cpu}, overhead={self.overhead}, '
                  f'cost_writes={self.cost_write}, cost_calc={self.cost_calc}')
        n_missing = len(missing)
        cost = self.cost_write + self.cost_calc
        limit = cost / self.cost_write
        log.debug(f'Limit on parallel workers from database contention is {limit:3.1f}')
        log.debug(f'Limit on parallel workers from CPU count is {n_cpu:d}')
        n_parallel = int(min(limit, n_cpu))
        n_total = int((n_missing + n_parallel - 1) / n_parallel)
        log.debug(f'Limit on total workers from work available is {n_total:d}')
        limit = cost * n_missing / self.overhead
        log.debug(f'Limit on total workers from overhead is {limit:3.1f}')
        n_total = min(n_total, int(limit))
        limit = n_parallel * MAX_REPEAT
        log.debug(f'Limit on total workers to boost batch size is {limit:d}')
        n_total = min(n_total, limit)
        log.info(f'Threads: {n_total}/{n_parallel}')
        return n_total, n_parallel

    def __spawn(self, s, missing, n_total, n_parallel):

        # unfortunately we have to do things with contiguous dates, which may introduce systematic
        # errors in our timing estimates

        n_missing = len(missing)
        workers = Workers(s, n_parallel, self.owner_out, self._base_command())
        start, finish = None, -1
        for i in range(n_total):
            start = finish + 1
            finish = int(0.5 + (i+1) * (n_missing-1) / n_total)
            if start > finish: raise Exception('Bad chunking logic')
            workers.run(self._args(missing, start, finish))
        workers.wait()

    # as a general rule, _missing and _args should be implemented together
    @abstractmethod
    def _args(self, missing, start, finish):
        raise NotImplementedError()

    @abstractmethod
    def _base_command(self):
        raise NotImplementedError()


class UniProcPipeline(MultiProcPipeline):

    def __init__(self, *args, overhead=None, cost_calc=None, cost_write=None, n_cpu=None, worker=None, id=None,
                 **kargs):
        super().__init__(*args, overhead=1, cost_calc=0, cost_write=1, n_cpu=None, worker=None, id=None, **kargs)


class LoaderMixin:

    def _get_loader(self, s, **kargs):
        if 'owner' not in kargs:
            kargs['owner'] = self.owner_out
        return StatisticJournalLoader(s, **kargs)




from urwid import WEIGHT, connect_signal, Padding, Divider

from .fixed import Fixed
from .tabs import TabList, TabNode
from .widgets import DividedPile, SquareButton
from ...lib.widgets import App
from ...squeal.binders import Binder


class Editor(TabNode):
    '''
    An editor for simple database classes (Injury, ActivityGroup)
    '''

    # we have to be careful to work well with sqlalchemy's session semantics:
    # - general editing is done within a single session
    # - this includes reset, delete and adding empty new values
    #   (all can be done without session commit)
    # - to do this, we must keep a store of the windgets / binders
    #   so that we don't need to query (after initial loading)
    # - data are saved / discarded on final exit (only)

    def __init__(self, log, session, bar, Widget, Model):
        self.__log = log
        self.__session = session
        self.__bar = bar
        self.__Widget = Widget
        self.__Model = Model
        tabs = TabList()
        body = []
        for instance in self.__session.query(Model).order_by(Model.sort).all():
            widget = Widget(self.__log, tabs, self.__bar, self)
            widget.connect(Binder(self.__log, self.__session, widget, instance=instance))
            body.append(widget)
        # and a button to add blanks
        more = SquareButton('More')
        body.append(tabs.append(Padding(Fixed(more, 8), width='clip')))
        connect_signal(more, 'click', self.__add_blank)
        super().__init__(log, DividedPile(body), tabs)

    def __add_blank(self, _unused_widget):
        tabs = TabList()
        widget = self.__Widget(self.__log, tabs, self.__bar, self)
        widget.connect(Binder(self.__log, self.__session, widget, self.__Model))
        body = self._w.contents
        n = len(body)
        body.insert(n-1, (Divider(), (WEIGHT, 1)))
        body.insert(n-1, (widget, (WEIGHT, 1)))
        self._w.contents = body
        n = len(self)
        self.insert_many(n - 1, tabs)

    def remove(self, widget):
        body = self._w.contents
        index = list(map(lambda x: x[0], body)).index(widget)
        self.__log.debug('Index %d: %s' % (index, body[index]))
        del body[index]
        del body[index]
        self._w.contents = body
        self.discover(discard=True)


class EditorApp(App):

    def __init__(self, log, session, bar, name, Widget, Model):
        self.__session = session
        tabs = TabList()
        self.editor = tabs.append(Editor(log, session, bar, Widget, Model))
        super().__init__(log, name, bar, self.editor, tabs, session)

